/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/



/*
 * usercall.h
 *
 * Interaction between the module device driver and the user-level wrapper
 */

#ifndef _USERCALL_H_
#define _USERCALL_H_

#define INCLUDE_ALLOW_USERLEVEL
#define INCLUDE_ALLOW_VMM
#define INCLUDE_ALLOW_VMKERNEL
#define INCLUDE_ALLOW_MODULE
#include "includeCheck.h"

#ifdef DEBUG_STUB
#include "x86.h"
#endif

/*
 * User calls
 *
 * Don't change these without recompiling the monitor.
 */
 
typedef enum UserCallOperation { 
   USERCALL_NONE = 300,

#define UC(call) USERCALL_##call,
#include "usercalltable.h"

   USERCALL_LAST
} UserCallOperation;

#ifdef VMX86_LOG
extern const char *userCallNames[];
#endif

typedef struct RPCBlock {
  uint32 spaceHolder;  /* Currently not used for anything  */
} RPCBlock;

#define RPC_MAX_SIZE 256u


#ifdef MONITOR
extern void UserRPC(UserCallOperation op);
#endif

#ifdef USERLEVEL
typedef size_t (*UserHandler)(void *clientData, struct RPCBlock *rpcBlock);
#endif

typedef struct LogBufferRPCBlock {
   RPCBlock rpc;
   const char *location;
   char text[1];  // undefined length
} LogBufferRPCBlock;

#define LOG_BUFFER_SIZE (RPC_MAX_SIZE - sizeof(char *))

#ifdef DEBUG_STUB
typedef struct MonitorBreakpointRPCBlock {
   RPCBlock rpc;
   Task task;
   Bool handled;
} MonitorBreakpointRPCBlock;
#endif

#endif // ifndef _USERCALL_H_
