/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/

#ifndef _X86CPUID_H_
#define _X86CPUID_H_

/* http://www.sandpile.org/ia32/cpuid.htm */

/* Common CPUID feature bits */
#define CPUID_FEATURE_COMMON_FPU    0x00000001 /* 0 */
#define CPUID_FEATURE_COMMON_VME    0x00000002 /* 1 */
#define CPUID_FEATURE_COMMON_DBGE   0x00000004 /* 2 */
#define CPUID_FEATURE_COMMON_PGSZE  0x00000008 /* 3 */
#define CPUID_FEATURE_COMMON_TSC    0x00000010 /* 4 */
#define CPUID_FEATURE_COMMON_MSR    0x00000020 /* 5 */
#define CPUID_FEATURE_COMMON_PAE    0x00000040 /* 6 */ 
#define CPUID_FEATURE_COMMON_MCK    0x00000080 /* 7 */ 
#define CPUID_FEATURE_COMMON_CPMX   0x00000100 /* 8 */
#define CPUID_FEATURE_COMMON_APIC   0x00000200 /* 9 */
#define CPUID_FEATURE_COMMON_RSVD10 0x00000400 /* 10 */
#define CPUID_FEATURE_COMMON_SEP    0x00000800 /* 11 */
#define CPUID_FEATURE_COMMON_MTTR   0x00001000 /* 12 */
#define CPUID_FEATURE_COMMON_PGE    0x00002000 /* 13 */
#define CPUID_FEATURE_COMMON_MCA    0x00004000 /* 14 */
#define CPUID_FEATURE_COMMON_CMOV   0x00008000 /* 15 */
#define CPUID_FEATURE_COMMON_PAT    0x00010000 /* 16 */
#define CPUID_FEATURE_COMMON_36PG   0x00020000 /* 17 */
#define CPUID_FEATURE_COMMON_CLFL   0x00080000 /* 19 */
#define CPUID_FEATURE_COMMON_MMX    0x00800000 /* 23 */
#define CPUID_FEATURE_COMMON_FXSAVE 0x01000000 /* 24 */
#define CPUID_FEATURE_COMMON_XMM    0x02000000 /* 25 */
#define CPUID_FEATURE_COMMON_XMM2   0x04000000 /* 26 */

/* Intel CPUID feature bits */
#define CPUID_FEATURE_INTEL_PSN    0x00040000
#define CPUID_FEATURE_INTEL_RSVD20 0x00100000
#define CPUID_FEATURE_INTEL_DTES   0x00200000
#define CPUID_FEATURE_INTEL_ACPI   0x00400000
/* CPUID_FEATURE_COMMON_MMX    */
/* CPUID_FEATURE_COMMON_FXSAVE */
/* CPUID_FEATURE_COMMON_XMM    */
/* CPUID_FEATURE_COMMON_XMM2   */
#define CPUID_FEATURE_INTEL_SS     0x08000000 
#define CPUID_FEATURE_INTEL_HT     0x10000000  // Pentium 4 sets it to 1!
#define CPUID_FEATURE_INTEL_TM     0x20000000 
#define CPUID_FEATURE_INTEL_RSVD30 0x40000000 
#define CPUID_FEATURE_INTEL_PBE    0x80000000

#define CPUID_FEATURE_INTEL_RSVD		\
        (  CPUID_FEATURE_COMMON_RSVD10		\
         | CPUID_FEATURE_INTEL_RSVD20		\
	 | CPUID_FEATURE_INTEL_RSVD30)

#define CPUID_EXTENDED_FEATURE_INTEL_PNI       0x00000001 /* 0 */
#define CPUID_EXTENDED_FEATURE_INTEL_MWAIT     0x00000008 /* 3 */
#define CPUID_EXTENDED_FEATURE_INTEL_DSCPL     0x00000010 /* 4 */
#define CPUID_EXTENDED_FEATURE_INTEL_TM2       0x00000100 /* 8 */
#define CPUID_EXTENDED_FEATURE_INTEL_HTCACHE   0x00000400 /* 10 */
#define CPUID_EXTENDED_FEATURE_INTEL_xPPR      0x00004000 /* 14 */

/* AMD CPUID feature bits */
#define CPUID_FEATURE_AMD_RSVD18  0X00040000
/* CPUID_FEATURE_COMMON_CLFL   */
#define CPUID_FEATURE_AMD_RSVD20  0X00100000
#define CPUID_FEATURE_AMD_RSVD21  0X00200000
#define CPUID_FEATURE_AMD_MMX_EXT 0X00400000
/* CPUID_FEATURE_COMMON_MMX    */
/* CPUID_FEATURE_COMMON_FXSAVE */
/* CPUID_FEATURE_COMMON_XMM    */
/* CPUID_FEATURE_COMMON_XMM2   */
#define CPUID_FEATURE_AMD_RSVD27 0X08000000
#define CPUID_FEATURE_AMD_RSVD28 0X10000000
#define CPUID_FEATURE_AMD_EXTENDED_NX        0X00100000
#define CPUID_FEATURE_AMD_EXTENDED_LM        0X20000000
#define CPUID_FEATURE_AMD_EXTENDED_3DNOWPLUS 0X40000000
#define CPUID_FEATURE_AMD_EXTENDED_3DNOW     0X80000000

#define CPUID_FEATURE_AMD_RSVD 			\
        (  CPUID_FEATURE_COMMON_RSVD10 		\
         | CPUID_FEATURE_AMD_RSVD18		\
         | CPUID_FEATURE_AMD_RSVD20		\
         | CPUID_FEATURE_AMD_RSVD21		\
         | CPUID_FEATURE_AMD_RSVD27		\
         | CPUID_FEATURE_AMD_RSVD28)

#define CPUID_STEPPING_MASK  0xf
#define CPUID_STEPPING(_eax) ((_eax) & CPUID_STEPPING_MASK)

#define CPUID_FAMILY(_eax)   (((_eax) >> 8) & 0xf)

/* Intel CPU Family */
#define CPUID_FAMILY_486     4
#define CPUID_FAMILY_P5      5
#define CPUID_FAMILY_P6      6
#define CPUID_FAMILY_EXTENDED 15

/* AMD CPU Family */
#define CPUID_FAMILY_5x86    4
#define CPUID_FAMILY_K5      5
#define CPUID_FAMILY_K6      5
#define CPUID_FAMILY_K7      6

#define CPUID_EXTENDED_FAMILY(_eax)   (((_eax) >> 20) & 0xff)

#define CPUID_EXTENDED_FAMILY_PENTIUM4  0
#define CPUID_EXTENDED_FAMILY_OPTERON   0

/*
 * Notice that CPUID families for Intel and AMD overlap. The following macros
 * should only be used AFTER the manufacturer has been established (through
 * the use of CPUID standard function 0).
 */
#define CPUID_FAMILY_IS_486(_eax) (CPUID_FAMILY(_eax) == CPUID_FAMILY_486)
#define CPUID_FAMILY_IS_P5(_eax)  (CPUID_FAMILY(_eax) == CPUID_FAMILY_P5)
#define CPUID_FAMILY_IS_P6(_eax)  (CPUID_FAMILY(_eax) == CPUID_FAMILY_P6)
#define CPUID_FAMILY_IS_PENTIUM4(_eax)  \
           (CPUID_FAMILY(_eax) == CPUID_FAMILY_EXTENDED && \
            CPUID_EXTENDED_FAMILY(_eax) == CPUID_EXTENDED_FAMILY_PENTIUM4)

#define CPUID_FAMILY_IS_OPTERON(_eax) \
           (CPUID_FAMILY(_eax) == CPUID_FAMILY_EXTENDED && \
            CPUID_EXTENDED_FAMILY(_eax) == CPUID_EXTENDED_FAMILY_OPTERON)

#define CPUID_MODEL(_eax)    (((_eax) >> 4) & 0xf)

#define CPUID_MODEL_PPRO       1
#define CPUID_MODEL_PII_3      3
#define CPUID_MODEL_PII_5      5
#define CPUID_MODEL_CELERON_6  6
#define CPUID_MODEL_PIII_7     7
#define CPUID_MODEL_PIII_8     8
#define CPUID_MODEL_PIII_A     10
#define CPUID_MODEL_EXTENDED   15

#define CPUID_EXTENDED_MODEL(_eax)    (((_eax) >> 16) & 0xf)

#define CPUID_TYPE(_eax)    (((_eax) >> 12) & 0x3)
#define CPUID_TYPE_PRIMARY     0
#define CPUID_TYPE_OVERDRIVE   1
#define CPUID_TYPE_SECONDARY   2

#endif
