/* **********************************************************
 * Copyright 2003 VMware, Inc.  All rights reserved.
 * -- VMware Confidential
 * **********************************************************/

/*
 * x86types.h --
 *
 *	Type definitions for the x86 architecture.
 */

#ifndef _X86TYPES_H_
#define _X86TYPES_H_

#define INCLUDE_ALLOW_USERLEVEL
#define INCLUDE_ALLOW_VMM
#define INCLUDE_ALLOW_MODULE
#define INCLUDE_ALLOW_VMK_MODULE
#define INCLUDE_ALLOW_VMKERNEL
#define INCLUDE_ALLOW_DISTRIBUTE
#include "includeCheck.h"


/*
 * Virtual, physical, machine address and page conversion macros
 *
 * XXX These should either go in vm_basic_types.h, where the types
 * are defined, or those definitions should be moved here.
 */

#define VA_2_VPN(_va)  ((_va) >> PAGE_SHIFT)
#define PTR_2_VPN(_ptr) VA_2_VPN((VA)(_ptr))
#define VPN_2_VA(_vpn) ((_vpn) << PAGE_SHIFT)

/*
 * Notice that we don't cast PA_2_PPN's argument to an unsigned type, because
 * we would lose compile-time checks for pointer operands and byte-sized
 * operands. If you use a signed quantity for _pa, ones may be shifted into the
 * high bits of your ppn.
 */

#define PA_2_PPN(_pa)  ((_pa) >> PAGE_SHIFT)
#define PPN_2_PA(_ppn)    ((_ppn) << PAGE_SHIFT)

static INLINE MA  MPN_2_MA(MPN mpn) { return (MA)mpn << PAGE_SHIFT; }
static INLINE MPN MA_2_MPN(MA  ma)  { return (MPN)(ma >> PAGE_SHIFT); }


/*
 * Linear address
 */

typedef uint32 LA;
typedef uint32 LPN;
#define LA_2_LPN(_la)     ((_la)>>PAGE_SHIFT)
#define LPN_2_LA(_lpn)    ((_lpn) << PAGE_SHIFT)

#define LAST_LPN ((1<<(8 * sizeof(LA) - PAGE_SHIFT)) - 1)

/* Valid bits in a LPN. */
#define LPN_MASK LAST_LPN


/*
 * Types used for PL4 page table in x86_64
 */

typedef uint64 VM_PL4_PML4E;
typedef uint64 VM_PL4_PDPE;
typedef uint64 VM_PL4_PDE;
typedef uint64 VM_PL4_PTE;


/*
 * 4 Mb pages
 */

#define VM_LARGE_PAGE_SHIFT  22
#define VM_LARGE_PAGE_SIZE   (1 << VM_LARGE_PAGE_SHIFT)
#define VM_LARGE_PAGE_MASK   (VM_LARGE_PAGE_SIZE - 1)


/*
 * Page table
 */

typedef uint32 VM_PDE;
typedef uint32 VM_PTE;
typedef uint64 VM_PAE_PDE;
typedef uint64 VM_PAE_PTE;
typedef uint64 VM_PDPTE;


typedef int32  Reg;


#endif // ifndef _X86TYPES_H_
