#ifndef __COMPAT_WAIT_H__
#   define __COMPAT_WAIT_H__


#include <linux/wait.h>
#include <linux/poll.h>
#include <linux/file.h>

#include "compat_file.h"


/*
 * The DECLARE_WAITQUEUE() API appeared in 2.3.1
 * It was back ported in 2.2.18
 *
 *  --hpreg
 */

#ifndef DECLARE_WAITQUEUE

typedef struct wait_queue *wait_queue_head_t;
#   define init_waitqueue_head(_headPtr) *(_headPtr) = NULL
#   define DECLARE_WAITQUEUE(_var, _task) \
   struct wait_queue _var = {_task, NULL, }

typedef struct wait_queue wait_queue_t;
#   define init_waitqueue_entry(_wait, _task) ((_wait)->task = (_task))

#endif

/*
 * The 'struct poll_wqueues' appeared in 2.5.48, when global
 * /dev/epoll interface was added.  It was backported to the
 * 2.4.20-wolk4.0s.
 */

#if VMW_HAVE_EPOLL // {
#define compat_poll_wqueues struct poll_wqueues
#else // } {
#define compat_poll_wqueues poll_table
#endif // }

#if VMW_HAVE_EPOLL // {

/* If prototype does not match, build will abort here */
extern void poll_initwait(compat_poll_wqueues *);

#define compat_poll_initwait(wait, table) ( \
   poll_initwait((table)), \
   (wait) = &(table)->pt \
)

#define compat_poll_freewait(wait, table) ( \
   poll_freewait((table)) \
)

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 0) // {

/* If prototype does not match, build will abort here */
extern void poll_initwait(compat_poll_wqueues *);

#define compat_poll_initwait(wait, table) ( \
   (wait) = (table), \
   poll_initwait(wait) \
)

#define compat_poll_freewait(wait, table) ( \
   poll_freewait((table)) \
)

#else // } {

#define compat_poll_initwait(wait, table) ( \
   (wait) = (table), /* confuse compiler */ \
   (wait) = (poll_table *) __get_free_page(GFP_KERNEL), \
   (wait)->nr = 0, \
   (wait)->entry = (struct poll_table_entry *)((wait) + 1), \
   (wait)->next = NULL \
)

static INLINE void
poll_freewait(poll_table *wait)
{
   while (wait) {
      struct poll_table_entry * entry;
      poll_table *old;

      entry = wait->entry + wait->nr;
      while (wait->nr > 0) {
	 wait->nr--;
	 entry--;
	 remove_wait_queue(entry->wait_address, &entry->wait);
	 compat_fput(entry->filp);
      }
      old = wait;
      wait = wait->next;
      free_page((unsigned long) old);
   }
}

#define compat_poll_freewait(wait, table) ( \
   poll_freewait((wait)) \
)

#endif // }


#endif /* __COMPAT_WAIT_H__ */
